@@ -1,3 +1,4 @@
+#![allow(non_snake_case)]
 use std::{
     io::{BufRead, BufReader, Read},
     ops::{Range, RangeFrom, RangeFull},
@@ -8,9 +9,9 @@ use nom::{
     branch::alt,
     bytes::streaming::{escaped, is_not, tag, take_while, take_while1},
     character::{
-        complete::{alphanumeric1 as alphanumeric, char, none_of, one_of},
+        complete::{alphanumeric1 as alphanumeric, char, multispace1, none_of, one_of},
         is_digit, is_hex_digit,
-        streaming::multispace0,
+        streaming::{alpha1, alphanumeric1, digit1, multispace0},
     },
     combinator::{cut, map, opt, recognize, value},
     error::{
@@ -80,7 +81,8 @@ fn utf8_char_width(b: u8) -> usize {
 // [2] Char ::= #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]
 #[inline]
 pub fn is_xml_char_t(chr: char) -> bool {
-    (chr >= '\u{A}' && chr <= '\u{D}')
+    chr == '\u{9}'
+        || (chr >= '\u{A}' && chr <= '\u{D}')
         || (chr >= '\u{20}' && chr <= '\u{D7FF}')
         || (chr >= '\u{E000}' && chr <= '\u{FFFD}')
         || (chr >= '\u{10000}' && chr <= '\u{10FFFF}')
@@ -196,6 +198,7 @@ where
             match f.parse(i.clone()) {
                 Err(Err::Error(_)) => return Ok((i, ())),
                 // Err(e) => return Err(e),
+                // ref#streamcut
                 Err(e) => return Ok((i, ())),
                 Ok((i1, o)) => {
                     // infinite loop check: the parser must always consume
@@ -444,10 +447,10 @@ fn test_stag() {
 
 // [14] CharData ::= [^<&]* - ([^<&]* ']]>' [^<&]*)
 // no '>' except ']]>'
-
+// The spec is not clear but we also apply Char restrictions
 #[inline]
 pub fn is_CharData_single_pure_t(chr: char) -> bool {
-    chr != '<' && chr != '&'
+    chr != '<' && chr != '&' && is_xml_char_t(chr)
 }
 
 pub fn CharData_single_pure(input: &[u8]) -> IResult<&[u8], &[u8]> {
@@ -590,6 +593,8 @@ enum ContentRelaxed<'a> {
     EmptyElemTag(StartElement<'a>),
     EndElement(EndElement<'a>),
     Reference(Reference<'a>),
+    CdataStart,
+    CommentStart,
 }
 
 fn content_relaxed_CharData(input: &[u8]) -> IResult<&[u8], ContentRelaxed> {
@@ -632,7 +637,21 @@ fn content_relaxed_Reference(input: &[u8]) -> IResult<&[u8], ContentRelaxed> {
     }
 }
 
-// [custom] relaxed ::= CharData | STag | EmptyElemTag | ETag | ... todo: add
+fn content_relaxed_CdataStart(input: &[u8]) -> IResult<&[u8], ContentRelaxed> {
+    match CDATASection_start(input) {
+        Ok(succ) => Ok((succ.0, ContentRelaxed::CdataStart)),
+        Err(err) => return Err(err),
+    }
+}
+
+fn content_relaxed_CommentStart(input: &[u8]) -> IResult<&[u8], ContentRelaxed> {
+    match Comment_start(input) {
+        Ok(succ) => Ok((succ.0, ContentRelaxed::CommentStart)),
+        Err(err) => return Err(err),
+    }
+}
+
+// [custom] relaxed ::= CharData | STag | EmptyElemTag | ETag | Reference | CDATA | Comment ... todo: add PI
 fn content_relaxed(input: &[u8]) -> IResult<&[u8], ContentRelaxed> {
     alt((
         content_relaxed_CharData,
@@ -640,6 +659,8 @@ fn content_relaxed(input: &[u8]) -> IResult<&[u8], ContentRelaxed> {
         content_relaxed_EmptyElemTag,
         content_relaxed_ETag,
         content_relaxed_Reference,
+        content_relaxed_CdataStart,
+        content_relaxed_CommentStart,
     ))(input)
 }
 
@@ -655,9 +676,348 @@ fn test_xml3() {
     println!("{:?}", res);
 }
 
+// Parser Rules organized by W3C Spec
+
+// [26] VersionNum ::= '1.' [0-9]+
+fn VersionNum(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((tag("1."), digit1)))(input)
+}
+
+#[test]
+fn test_VersionNum() {
+    let data = r#"1.123 "#.as_bytes();
+    let res = VersionNum(&data);
+    println!("{:?}", res);
+}
+//  [24]   	VersionInfo	   ::=   	S 'version' Eq ("'" VersionNum "'" | '"' VersionNum '"')
+
+fn VersionInfo(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((
+        multispace1,
+        tag("version"),
+        Eq,
+        alt((
+            delimited(char('"'), VersionNum, char('"')),
+            delimited(char('\''), VersionNum, char('\'')),
+        )),
+    )))(input)
+}
+#[test]
+fn test_VersionInfo() {
+    let data = r#"  version="1.0" "#.as_bytes();
+    let res = VersionInfo(&data);
+    println!("{:?}", res);
+}
+
+// [81]   	EncName	   ::=   	[A-Za-z] ([A-Za-z0-9._] | '-')*
+fn EncName(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((
+        alpha1,
+        many0_custom_trycomplete(alt((alphanumeric1, tag("-"), tag("."), tag("_")))),
+    )))(input)
+}
+#[test]
+fn test_EncName() {
+    let data = r#"UTF-8 "#.as_bytes();
+    let res = EncName(&data);
+    println!("{:?}", res);
+}
+
+// [80]   	EncodingDecl	   ::=   	S 'encoding' Eq ('"' EncName '"' | "'" EncName "'" )
+fn EncodingDecl(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((
+        multispace1,
+        tag("encoding"),
+        Eq,
+        alt((
+            delimited(char('"'), EncName, char('"')),
+            delimited(char('\''), EncName, char('\'')),
+        )),
+    )))(input)
+}
+#[test]
+fn test_EncodingDecl() {
+    let data = r#" encoding='EUC-JP' "#.as_bytes();
+    let res = EncodingDecl(&data);
+    println!("{:?}", res);
+}
+
+// [32] SDDecl ::= S 'standalone' Eq (("'" ('yes' | 'no') "'") | ('"' ('yes' | 'no') '"'))
+fn yes_mi_no_mu(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    alt((tag("yes"), tag("no")))(input)
+}
+fn SDDecl(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((
+        multispace1,
+        tag("standalone"),
+        Eq,
+        alt((
+            delimited(char('"'), yes_mi_no_mu, char('"')),
+            delimited(char('\''), yes_mi_no_mu, char('\'')),
+        )),
+    )))(input)
+}
+#[test]
+fn test_SDDecl() {
+    let data = r#"  standalone='yes' "#.as_bytes();
+    let res = SDDecl(&data);
+    println!("{:?}", res);
+}
+
+// [23] XMLDecl ::= '<?xml' VersionInfo EncodingDecl? SDDecl? S? '?>'
+fn XMLDecl(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((
+        tag("<?xml"),
+        VersionInfo,
+        opt(EncodingDecl),
+        opt(SDDecl),
+        multispace0,
+        tag("?>"),
+    )))(input)
+}
+
+// [27]   	Misc	   ::=   	Comment | PI | S
+//todo: comment | PI, we may need to separate
+fn Misc(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(alt((multispace1,)))(input)
+}
+
+fn docstart_custom(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((XMLDecl, multispace0)))(input)
+}
+
+#[test]
+fn test_XMLDecl() {
+    let data = r#"<?xml version="1.0"  encoding="UTF-8" standalone='yes'?>"#.as_bytes();
+    let res = XMLDecl(&data);
+    println!("{:?}", res);
+}
+
+// [1] document ::= prolog element Misc*
+// [22]   	prolog	   ::=   	XMLDecl? Misc* (doctypedecl Misc*)?
+
+// [15] Comment ::= '<!--' ((Char - '-') | ('-' (Char - '-')))* '-->'
+//spec seems to not allow empty comments? There are parsers that allow it.
+fn Comment_start(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    tag("<!--")(input)
+}
+
+fn Comment_end(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    tag("-->")(input)
+}
+
+// We don't need to exclude "-" we handle that in inside_Comment_single
+// #[inline]
+//  fn is_CharData_single_pure_t(chr: char) -> bool {
+//     chr != '<' && chr != '&' && is_xml_char_t(chr)
+// }
+
+fn inside_Comment_or_CDATA_single_pure(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    if input.len() == 0 {
+        return Err(Err::Incomplete(Needed::new(1)));
+    }
+    let width = utf8_char_width(input[0]);
+
+    if input.len() < width {
+        return Err(Err::Incomplete(Needed::new(width - input.len())));
+    }
+
+    let c = match std::str::from_utf8(&input[..width]).ok() {
+        Some(s) => s.chars().next().unwrap(),
+        None => return Err(Err::Error(Error::new(input, ErrorKind::Char))),
+    };
+
+    if is_xml_char_t(c) {
+        return Ok((&input[width..], &input[0..width]));
+    } else {
+        return Err(Err::Error(Error::new(input, ErrorKind::Char)));
+    }
+}
+
+fn inside_Comment_single(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    //if input = 0 , don't send incomplete
+    // ref#streamcut
+    if input.len() == 0 {
+        return Err(Err::Error(Error::new(input, ErrorKind::Char)));
+    }
+
+    // '--' should not appear in the comment, if we can't be sure because input is eof, we should request more data.
+    match tag::<&str, &[u8], Error<&[u8]>>("--")(input) {
+        Ok(r) => return Err(Err::Error(Error::new(input, ErrorKind::Char))),
+        Err(Err::Incomplete(n)) => return Err(Err::Incomplete(Needed::new(1))),
+        _ => (),
+    };
+    inside_Comment_or_CDATA_single_pure(input)
+}
+
+fn Comment(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((
+        Comment_start,
+        many0_custom_chardata(inside_Comment_single),
+        Comment_end,
+    )))(input)
+}
+
+#[test]
+fn test_comment() {
+    assert_eq!(
+        Comment("<!-- comment -->a".as_bytes()),
+        Ok((&b"a"[..], &b"<!-- comment -->"[..]))
+    );
+
+    assert_eq!(
+        Comment("<!---->cc".as_bytes()),
+        Ok((&b"cc"[..], &b"<!---->"[..]))
+    );
+
+    assert_eq!(
+        Comment("<!-- comment --->a".as_bytes()),
+        Err(Err::Error(error_position!(
+            "--->a".as_bytes(),
+            ErrorKind::Tag
+        )))
+    );
+
+    assert_eq!(
+        Comment("<!-- com--ment -->a".as_bytes()),
+        Err(Err::Error(error_position!(
+            "--ment -->a".as_bytes(),
+            ErrorKind::Tag
+        )))
+    );
+
+    assert_eq!(
+        Comment("<!--ok-".as_bytes()),
+        Err(Err::Incomplete(Needed::new(2)))
+    );
+    assert_eq!(
+        Comment("<!--ok--".as_bytes()),
+        Err(Err::Incomplete(Needed::new(1)))
+    );
+}
+
+enum InsideComment<'a> {
+    Characters(&'a [u8]),
+    CommentEnd,
+}
+
+fn insidecomment_characters(input: &[u8]) -> IResult<&[u8], InsideComment> {
+    match recognize(tuple((
+        inside_Comment_single,
+        many0_custom_chardata(inside_Comment_single),
+    )))(input)
+    {
+        Ok(succ) => Ok((succ.0, InsideComment::Characters(succ.1))),
+        Err(err) => return Err(err),
+    }
+}
+
+fn insidecomment_comment_end(input: &[u8]) -> IResult<&[u8], InsideComment> {
+    match Comment_end(input) {
+        Ok(succ) => Ok((succ.0, InsideComment::CommentEnd)),
+        Err(err) => return Err(err),
+    }
+}
+
+// [custom]
+fn insidecomment(input: &[u8]) -> IResult<&[u8], InsideComment> {
+    alt((insidecomment_characters, insidecomment_comment_end))(input)
+}
+
+// [18] CDSect ::= CDStart CData CDEnd
+
+// [19] CDStart ::= '<![CDATA['
+fn CDATASection_start(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    tag("<![CDATA[")(input)
+}
+// [21] CDEnd ::= ']]>'
+fn CDATASection_end(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    tag("]]>")(input)
+}
+
+// [20] CData ::= (Char* - (Char* ']]>' Char*))
+
+fn inside_CDATASection_single(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    //if input = 0 , don't send incomplete
+    // ref#streamcut
+    if input.len() == 0 {
+        return Err(Err::Error(Error::new(input, ErrorKind::Char)));
+    }
+
+    // ']]>' should not appear in the cdata section, if we can't be sure because input is eof, we should request more data.
+    match tag::<&str, &[u8], Error<&[u8]>>("]]>")(input) {
+        Ok(r) => return Err(Err::Error(Error::new(input, ErrorKind::Char))),
+        Err(Err::Incomplete(n)) => return Err(Err::Incomplete(Needed::Unknown)),
+        _ => (),
+    };
+    inside_Comment_or_CDATA_single_pure(input)
+}
+
+fn CDATASection(input: &[u8]) -> IResult<&[u8], &[u8]> {
+    recognize(tuple((
+        CDATASection_start,
+        many0_custom_chardata(inside_CDATASection_single),
+        CDATASection_end,
+    )))(input)
+}
+
+#[test]
+fn test_cdata() {
+    assert_eq!(
+        CDATASection("<![CDATA[abc]]>a".as_bytes()),
+        Ok((&b"a"[..], &b"<![CDATA[abc]]>"[..]))
+    );
+
+    assert_eq!(
+        CDATASection("<![CDATA[]]>".as_bytes()),
+        Ok((&b""[..], &b"<![CDATA[]]>"[..]))
+    );
+
+    assert_eq!(
+        CDATASection("<![CDATA[ ]]".as_bytes()),
+        Err(Err::Incomplete(Needed::new(1)))
+    );
+    assert_eq!(
+        CDATASection("<![CDATA[ ]".as_bytes()),
+        Err(Err::Incomplete(Needed::new(2)))
+    );
+}
+
+enum InsideCdata<'a> {
+    Characters(&'a [u8]),
+    CdataEnd,
+}
+
+fn insidecdata_characters(input: &[u8]) -> IResult<&[u8], InsideCdata> {
+    match recognize(tuple((
+        inside_CDATASection_single,
+        many0_custom_chardata(inside_CDATASection_single),
+    )))(input)
+    {
+        Ok(succ) => Ok((succ.0, InsideCdata::Characters(succ.1))),
+        Err(err) => return Err(err),
+    }
+}
+
+fn insidecdata_cdata_end(input: &[u8]) -> IResult<&[u8], InsideCdata> {
+    match CDATASection_end(input) {
+        Ok(succ) => Ok((succ.0, InsideCdata::CdataEnd)),
+        Err(err) => return Err(err),
+    }
+}
+
+// [custom]
+fn insidecdata(input: &[u8]) -> IResult<&[u8], InsideCdata> {
+    alt((insidecdata_characters, insidecdata_cdata_end))(input)
+}
+
+#[derive(Clone, Debug, Eq, PartialEq)]
 enum ParserState {
     Content,
-    TagStart,
+    DocStart,
+    DocEnd,
+    InsideCdata,
+    InsideComment,
 }
 
 pub struct OxideParser<R: Read> {
@@ -671,7 +1031,7 @@ pub struct OxideParser<R: Read> {
 fn convert_start_element<'a>(
     strbuffer: &'a mut String,
     event1: StartElement,
-) -> xml_sax::Event<'a> {
+) -> xml_sax::StartElement<'a> {
     let start = strbuffer.len();
     let size = event1.name.len();
     strbuffer.push_str(event1.name);
@@ -710,14 +1070,15 @@ fn convert_start_element<'a>(
 
         attributes.push(xml_sax::Attribute {
             value: &strbuffer[att.value],
-            qualified_name: &strbuffer[att.qualified_name],
+            name: &strbuffer[att.qualified_name],
         });
     }
 
-    xml_sax::Event::StartElement(xml_sax::StartElement {
+    xml_sax::StartElement {
         name: &strbuffer[start..(start + size)],
         attributes: attributes,
-    })
+        is_empty: false,
+    }
 }
 
 fn push_str_get_range(strbuffer: &mut String, addition: &str) -> Range<usize> {
@@ -737,7 +1098,7 @@ impl<R: Read> OxideParser<R> {
 
     pub fn start(reader: R) -> OxideParser<R> {
         OxideParser {
-            state: ParserState::Content,
+            state: ParserState::DocStart,
             bufreader: BufReader::with_capacity(8192, reader),
             offset: 0,
             buffer2: vec![],
@@ -770,92 +1131,178 @@ impl<R: Read> OxideParser<R> {
         //     attributes: vec![],
         // };
         // let mut event1: StartElement; //<'b>; //&'a
-        let mut event2: xml_sax::Event;
-
-        let res = content_relaxed(&self.buffer2);
-        match res {
-            Ok(parseresult) => {
-                self.offset = self.buffer2.offset(parseresult.0);
-
-                match parseresult.1 {
-                    ContentRelaxed::CharData(event1) => {
-                        let start = self.strbuffer.len();
-                        let size = event1.len();
-                        self.strbuffer
-                            .push_str(unsafe { std::str::from_utf8_unchecked(event1) });
-
-                        event2 = xml_sax::Event::Characters(&self.strbuffer[start..(start + size)])
+        let event2: xml_sax::Event;
+        match self.state {
+            ParserState::DocStart => {
+                let res = docstart_custom(&self.buffer2);
+                match res {
+                    Ok(parseresult) => {
+                        self.offset = self.buffer2.offset(parseresult.0);
+                        self.state = ParserState::Content;
+                        return xml_sax::Event::StartDocument;
                     }
-                    ContentRelaxed::StartElement(event1) => {
-                        //todo decode
-                        event2 = convert_start_element(&mut self.strbuffer, event1);
+                    Err(err) => panic!(),
+                }
+            }
+            ParserState::Content => {
+                let res = content_relaxed(&self.buffer2);
+                match res {
+                    Ok(parseresult) => {
+                        self.offset = self.buffer2.offset(parseresult.0);
+
+                        match parseresult.1 {
+                            ContentRelaxed::CharData(event1) => {
+                                let start = self.strbuffer.len();
+                                let size = event1.len();
+                                self.strbuffer
+                                    .push_str(unsafe { std::str::from_utf8_unchecked(event1) });
+
+                                event2 = xml_sax::Event::Characters(
+                                    &self.strbuffer[start..(start + size)],
+                                )
+                            }
+                            ContentRelaxed::StartElement(event1) => {
+                                //todo decode
+                                event2 = xml_sax::Event::StartElement(convert_start_element(
+                                    &mut self.strbuffer,
+                                    event1,
+                                ));
+                            }
+                            ContentRelaxed::EmptyElemTag(event1) => {
+                                //todo decode
+                                let mut start_elem =
+                                    convert_start_element(&mut self.strbuffer, event1);
+                                start_elem.is_empty = true;
+                                event2 = xml_sax::Event::StartElement(start_elem);
+                                //todo add endelement after this?
+                            }
+                            ContentRelaxed::EndElement(event1) => {
+                                let start = self.strbuffer.len();
+                                let size = event1.name.len();
+                                self.strbuffer.push_str(event1.name);
+
+                                event2 = xml_sax::Event::EndElement(xml_sax::EndElement {
+                                    name: &self.strbuffer[start..(start + size)],
+                                })
+                            }
+                            ContentRelaxed::Reference(event1) => {
+                                // let start = self.strbuffer.len();
+                                // let size = event1.initial.len();
+                                // let range_initial = Range {
+                                //     start: start,
+                                //     end: start + size,
+                                // };
+                                // self.strbuffer.push_str(event1.initial);
+
+                                let range: Range<usize> =
+                                    push_str_get_range(&mut self.strbuffer, event1.initial);
+
+                                let range_resolved = match event1.initial {
+                                    "&amp;" => push_str_get_range(&mut self.strbuffer, "&"),
+                                    "&lt" => push_str_get_range(&mut self.strbuffer, "<"),
+                                    "&gt;" => push_str_get_range(&mut self.strbuffer, ">"),
+                                    "&quot;" => push_str_get_range(&mut self.strbuffer, "\""),
+                                    "&apos;" => push_str_get_range(&mut self.strbuffer, "'"),
+                                    _ => push_str_get_range(&mut self.strbuffer, event1.initial),
+                                };
+
+                                //todo resolve char refs
+                                //we are ignoring DTD entity refs
+                                event2 = xml_sax::Event::Reference(xml_sax::Reference {
+                                    raw: &self.strbuffer[range],
+                                    resolved: &self.strbuffer[range_resolved],
+                                })
+                            }
+                            ContentRelaxed::CdataStart => {
+                                event2 = xml_sax::Event::StartCdata;
+                                self.state = ParserState::InsideCdata;
+                            }
+                            ContentRelaxed::CommentStart => {
+                                event2 = xml_sax::Event::StartComment;
+                                self.state = ParserState::InsideComment;
+                            }
+                        }
                     }
-                    ContentRelaxed::EmptyElemTag(event1) => {
-                        //todo decode
-                        event2 = convert_start_element(&mut self.strbuffer, event1);
-
-                        //todo add endelement after this?
+                    Err(Err::Incomplete(e)) => {
+                        // panic!()
+                        // self.read_data();
+                        // if read bytes are 0 then return eof, otherwise return dummy event
+                        if self.buffer2.len() == 0 {
+                            return xml_sax::Event::EndDocument;
+                        }
+                        println!("try to read bytes: {:?}", unsafe { &self.buffer2 });
+                        println!("try to read: {:?}", unsafe {
+                            std::str::from_utf8_unchecked(&self.buffer2)
+                        });
+                        println!("err: {:?}", e);
+                        panic!()
                     }
-                    ContentRelaxed::EndElement(event1) => {
-                        let start = self.strbuffer.len();
-                        let size = event1.name.len();
-                        self.strbuffer.push_str(event1.name);
+                    Err(e) => {
+                        println!("try to read bytes: {:?}", unsafe { &self.buffer2 });
+                        println!("try to read: {:?}", unsafe {
+                            std::str::from_utf8_unchecked(&self.buffer2)
+                        });
+                        println!("err: {:?}", e);
 
-                        event2 = xml_sax::Event::EndElement(xml_sax::EndElement {
-                            name: &self.strbuffer[start..(start + size)],
-                        })
-                    }
-                    ContentRelaxed::Reference(event1) => {
-                        // let start = self.strbuffer.len();
-                        // let size = event1.initial.len();
-                        // let range_initial = Range {
-                        //     start: start,
-                        //     end: start + size,
-                        // };
-                        // self.strbuffer.push_str(event1.initial);
-
-                        let range: Range<usize> =
-                            push_str_get_range(&mut self.strbuffer, event1.initial);
-
-                        let range_resolved = match event1.initial {
-                            "&amp;" => push_str_get_range(&mut self.strbuffer, "&"),
-                            "&lt" => push_str_get_range(&mut self.strbuffer, "<"),
-                            "&gt;" => push_str_get_range(&mut self.strbuffer, ">"),
-                            "&quot;" => push_str_get_range(&mut self.strbuffer, "\""),
-                            "&apos;" => push_str_get_range(&mut self.strbuffer, "'"),
-                            _ => push_str_get_range(&mut self.strbuffer, event1.initial),
-                        };
-
-                        //todo resolve char refs
-                        //we are ignoring DTD entity refs
-                        event2 = xml_sax::Event::Reference(xml_sax::Reference {
-                            original: &self.strbuffer[range],
-                            resolved: &self.strbuffer[range_resolved],
-                        })
+                        panic!()
                     }
                 }
             }
-            Err(Err::Incomplete(e)) => {
-                // panic!()
-                // self.read_data();
-                // if read bytes are 0 then return eof, otherwise return dummy event
-                println!("try to read bytes: {:?}", unsafe { &self.buffer2 });
-                println!("try to read: {:?}", unsafe {
-                    std::str::from_utf8_unchecked(&self.buffer2)
-                });
-                println!("err: {:?}", e);
-                panic!()
-            }
-            Err(e) => {
-                println!("try to read bytes: {:?}", unsafe { &self.buffer2 });
-                println!("try to read: {:?}", unsafe {
-                    std::str::from_utf8_unchecked(&self.buffer2)
-                });
-                println!("err: {:?}", e);
+            ParserState::DocEnd => todo!(),
+            ParserState::InsideCdata => {
+                //expect cdata or cdata-end
+                let res = insidecdata(&self.buffer2);
+                match res {
+                    Ok(parseresult) => {
+                        self.offset = self.buffer2.offset(parseresult.0);
+                        match parseresult.1 {
+                            InsideCdata::Characters(characters) => {
+                                let start = self.strbuffer.len();
+                                let size = characters.len();
+                                self.strbuffer
+                                    .push_str(unsafe { std::str::from_utf8_unchecked(characters) });
 
-                panic!()
+                                event2 = xml_sax::Event::Characters(
+                                    &self.strbuffer[start..(start + size)],
+                                )
+                            }
+                            InsideCdata::CdataEnd => {
+                                self.state = ParserState::Content;
+                                event2 = xml_sax::Event::EndCdata;
+                            }
+                        }
+                    }
+                    Err(err) => panic!(),
+                }
+            }
+            ParserState::InsideComment => {
+                //expect comment or comment-end
+                let res = insidecomment(&self.buffer2);
+                match res {
+                    Ok(parseresult) => {
+                        self.offset = self.buffer2.offset(parseresult.0);
+                        match parseresult.1 {
+                            InsideComment::Characters(characters) => {
+                                let start = self.strbuffer.len();
+                                let size = characters.len();
+                                self.strbuffer
+                                    .push_str(unsafe { std::str::from_utf8_unchecked(characters) });
+
+                                event2 = xml_sax::Event::Characters(
+                                    &self.strbuffer[start..(start + size)],
+                                )
+                            }
+                            InsideComment::CommentEnd => {
+                                self.state = ParserState::Content;
+                                event2 = xml_sax::Event::EndComment;
+                            }
+                        }
+                    }
+                    Err(err) => panic!(),
+                }
             }
         }
+
         event2
 
         // let res = STag(&self.buffer2);
@@ -898,6 +1345,7 @@ fn test_parser1() {
             xml_sax::Event::EndElement(_) => todo!(),
             xml_sax::Event::Characters(c) => {}
             xml_sax::Event::Reference(c) => {}
+            _ => {}
         }
     }
 
